# -*- coding: utf-8 -*-
"""
Created on Mon Feb 15 11:26:41 2016

@author: ih3
"""

import numpy

class State(object):
    """
    A state at a point. Initialized with the rest mass density, velocity, and
    specific internal energy, as well as an equation of state.
    """

    def __init__(self, rho, v, vt, eps, eos, label=None):
        r"""
        Constructor

        Parameters
        ----------

        rho : scalar
            Rest mass density :math:`\rho_0`
        v : scalar
            Velocity component in the normal (:math:`x`) direction :math:`v_x`
        v_t : scalar
            Velocity component tangential to :math:`x` :math:`v_t`
        eps : scalar
            Specific internal energy :math:`\epsilon`
        eos : dictionary
            Equation of State
        label : string
            Label for output purposes.
        """
        self.rho = rho
        self.v = v
        self.vt = vt
        self.eps = eps
        self.eos = eos
        if 'q_available' in self.eos:
            self.q = self.eos['q_available']
        else:
            self.q = None
        self.W_lorentz = 1.0 / numpy.sqrt(1.0 - self.v**2 - self.vt**2)
        self.p = self.eos['p_from_rho_eps'](rho, eps)
        self.h = self.eos['h_from_rho_eps'](rho, eps)
        self.cs = self.eos['cs_from_rho_eps'](rho, eps)
        self.label = label

    def prim(self):
        r"""
        Return the primitive variables :math:`\rho, v_x, v_t, \epsilon`.
        """
        return numpy.array([self.rho, self.v, self.vt, self.eps])

    def state(self):
        r"""
        Return all variables :math:`\rho, v_x, v_t, \epsilon, p, W, h, c_s`.
        """
        return numpy.array([self.rho, self.v, self.vt, self.eps, self.p,\
        self.W_lorentz, self.h, self.cs])

    def wavespeed(self, wavenumber):
        """
        Compute the wavespeed given the wave number (0 for the left wave,
        2 for the right wave).

        Parameters
        ----------

        wavenumber: scalar
            Wave number ([0,1,2]).
        """
        if wavenumber == 1:
            return self.v
        elif abs(wavenumber - 1) == 1:
            s = wavenumber - 1
            term1 = self.v * (1.0 - self.cs**2)
            term2 = (1.0 - self.v**2 - self.vt**2) * (1.0 - self.v**2 -
            self.vt**2 * self.cs**2)
            term3 = 1.0 - (self.v**2 + self.vt**2) * self.cs**2
            return (term1 + s * self.cs * numpy.sqrt(term2)) / term3
        else:
            raise NotImplementedError("wavenumber must be 0, 1, 2")

    def vt_from_known(self, rho, v, eps):
        r"""
        Computes tangential velocity across a wave.

        Parameters
        ----------

        self : State
            The known state
        rho : scalar
            Rest mass density in the unknown state across the wave
        v : scalar
            Normal velocity :math:`v_x` in the unknown state across the wave
        eps : scalar
            Specific internal energy in the unknown state across the wave

        Returns
        -------

        vt : scalar
            Tangential velocity :math:`v_t` in the unknown state
        """
        h = self.eos['h_from_rho_eps'](rho, eps)
        vt = self.h * self.W_lorentz * self.vt
        vt *= numpy.sqrt((1.0 - v**2)/
            (h**2 + (self.h * self.W_lorentz * self.vt)**2))
        return vt

    def latex_string(self):
        """
        Helper function to represent the state as a string.
        """
        if self.q:
            s = r"\begin{pmatrix} \rho \\ v_x \\ v_t \\ \epsilon \\ q \end{pmatrix}"
        else:
            s = r"\begin{pmatrix} \rho \\ v_x \\ v_t \\ \epsilon \end{pmatrix}"
        if self.label:
            s += r"_{{{}}} ".format(self.label)
        s += "= "
        if self.q:
            s += r"\begin{{pmatrix}} {:.4f} \\ {:.4f} \\ {:.4f} \\ {:.4f} \\ {:.4f} \end{{pmatrix}}".format(\
            self.rho, self.v, self.vt, self.eps, self.q)
        else:
            s += r"\begin{{pmatrix}} {:.4f} \\ {:.4f} \\ {:.4f} \\ {:.4f} \end{{pmatrix}}".format(\
            self.rho, self.v, self.vt, self.eps)
        return s

    def _repr_latex_(self):
        """
        IPython or Jupyter repr.
        """
        s = r"$" + self.latex_string() + r"$"
        return s



# -*- coding: utf-8 -*-
"""
Created on Tue Apr 26 09:48:54 2016

@author: ih3
"""

import numpy
from scipy.integrate import odeint
from scipy.optimize import brentq
from matplotlib.transforms import offset_copy
import matplotlib.transforms as mtrans
from . import wave

def rarefaction(ps, q_start, lr_sign):
    vs = numpy.zeros_like(ps)
    for i, p in enumerate(ps):
        w_all = odeint(wave.rarefaction_dwdp,
                       numpy.array([q_start.rho, q_start.v,
                       q_start.eps]),
                       [q_start.p, p], rtol = 1e-12,
                       atol = 1e-10,
                       args=((q_start, lr_sign+1)))
        vs[i] = w_all[-1, 1]
    return vs

def shock(ps, q_start, lr_sign):
    vs = numpy.zeros_like(ps)
    for i, p in enumerate(ps):
        j2, rho, eps, diffp = wave.mass_flux_squared(q_start,
                                    p, q_start.eos)
        v_shock, q_end = wave.post_discontinuity_state(p,
                                    q_start,
                                    lr_sign, "", j2,
                                    rho, eps, diffp)
        vs[i] = q_end.v
    return vs

def deflagration(ps, q_start, lr_sign):
    vs = numpy.zeros_like(ps)
    for i, p in enumerate(ps):
        j2, rho, eps, diffp = wave.mass_flux_squared(q_start,
                                    p, q_start.eos)
        v_deflagration, q_end = wave.post_discontinuity_state(p, q_start,
                                    lr_sign, "", j2,
                                    rho, eps, diffp)

        if (lr_sign*(q_end.wavespeed(lr_sign+1) - v_deflagration) < 0):
            p_cjdf = brentq(wave.deflagration_root,
                            (1.0+1e-9)*p,
                            (1.0-1e-9)*q_start.p,
                            args=(q_start, q_start.eos,
                            lr_sign+1, ""))
            j2, rho, eps, dp = wave.mass_flux_squared(q_start,
                            p_cjdf, q_start.eos)
            v_deflagration, q_end = wave.post_discontinuity_state(p_cjdf, q_start,
                                    lr_sign, "", j2,
                                    rho, eps, diffp)
        vs[i] = q_end.v
    return vs

def detonation(ps, q_start, lr_sign):
    vs = numpy.zeros_like(ps)
    for i, p in enumerate(ps):
        j2, rho, eps, diffp = wave.mass_flux_squared(q_start,
                                    p, q_start.eos)
        if j2 < 0:
            # The single detonation is unphysical - must be unstable weak
            # detonation. So skip the calculation and make sure the CJ
            # calculation runs
            q_unknown = deepcopy(q_start)
            v_detonation = q_unknown.wavespeed(lr_sign+1) + 1
        else:
            v_detonation, q_unknown = wave.post_discontinuity_state(p, q_start,
                                    lr_sign, "", j2,
                                    rho, eps, diffp,
                                    q_start.eos)

        if (lr_sign*(q_unknown.wavespeed(lr_sign+1) - v_detonation) < 0):
            pmin = (1.0+1e-9)*min(q_start.p, p)
            pmax = max(q_start.p, p)
            fmin = wave.deflagration_root(pmin, q_start, q_start.eos, lr_sign+1, "")
            fmax = wave.deflagration_root(pmax, q_start, q_start.eos, lr_sign+1, "")
            while fmin * fmax > 0:
                pmax *= 2.0
                fmax = wave.deflagration_root(pmax, q_start, q_start.eos, lr_sign+1, "")
            p_cjdt = brentq(wave.deflagration_root, pmin, pmax,
                            args=(q_start, q_start.eos, lr_sign+1, ""))
            j2, rho, eps, dp = wave.mass_flux_squared(q_start, p_cjdt, q_start.eos)
            v_detonation, q_unknown = wave.post_discontinuity_state(p_cjdt, q_start,
                                   lr_sign, "", j2,
                                   rho, eps, diffp,
                                   q_start.eos)
        vs[i] = q_unknown.v
    return vs

def find_pre_ignition(v_s, j2, rho, lr_sign):
    A = rho**2 / (j2 *
        (v_s - lr_sign * numpy.sqrt(1 + rho**2 / j2)))

    return -0.5 * A + numpy.sqrt(1 + A * v_s), \
           -0.5 * A - numpy.sqrt(1 + A * v_s)


def plot_P_v(rp, ax, fig, var_to_plot = "velocity"):
    """
    Plot the curves joining states within phase space for the Riemann Problem.

    Parameters
    ----------

    rp : RiemannProblem
        The Riemann Problem to be plotted
    ax : matplotlib axis
        The axis on which to plot
    var_to_plot : string
        The name of the variable to plot on the y axis
    """

    # TODO: This won't work with volume at the moment

    if var_to_plot == "velocity":
        var_index = 1
        var_invert = False
        var_name = r"$v$"
    elif var_to_plot == "volume":
        var_index = 0
        var_invert = True
        var_name = r"$V$"
    else:
        raise(ValueError, "var_to_plot ({}) not recognized".format(var_to_plot))
    p_min=min([rp.state_l.p, rp.state_r.p, rp.p_star])
    p_max=max([rp.state_l.p, rp.state_r.p, rp.p_star])
    if var_to_plot == "velocity":
        ax.plot(rp.state_l.v, rp.state_l.p, 'ko', label=r"$U_L$")
        ax.plot(rp.state_r.v, rp.state_r.p, 'k^', label=r"$U_R$")
        ax.plot(rp.state_star_l.v, rp.p_star, 'k*', label=r"$U_*$")
        trans_offset = mtrans.offset_copy(ax.transData, fig=fig, x=-15, y=-6, units='dots')
        ax.text(rp.state_l.v, rp.state_l.p, r"$U_L$", transform=trans_offset, horizontalalignment='center',
                 verticalalignment='bottom')
        trans_offset = mtrans.offset_copy(ax.transData, fig=fig, x=20, y=-6, units='dots')
        ax.text(rp.state_r.v, rp.state_r.p, r"$U_R$", transform=trans_offset, horizontalalignment='center',
                 verticalalignment='bottom')
        trans_offset = mtrans.offset_copy(ax.transData, fig=fig, x=20, y=-6, units='dots')
        ax.text(rp.state_star_l.v, rp.p_star, r"$U_*$", transform=trans_offset, horizontalalignment='center',
                 verticalalignment='bottom')
    else:
        ax.plot(1.0/rp.state_l.rho, rp.state_l.p, 'ko', label=r"$U_L$")
        ax.plot(1.0/rp.state_r.rho, rp.state_r.p, 'k^', label=r"$U_R$")
        ax.plot(1.0/rp.state_star_l.rho, rp.p_star, 'k*', label=r"$U_{*_L}$")
        ax.plot(1.0/rp.state_star_r.rho, rp.p_star, 'k<', label=r"$U_{*_R}$")
    dp = max(0.1, p_max-p_min)
    dp_fraction = min(0.5, 5*p_min)*dp

    p_l_1 = numpy.linspace(p_min-0.1*dp_fraction, rp.state_l.p-1e-3*dp_fraction)
    p_l_2 = numpy.linspace(rp.state_l.p+1e-3*dp_fraction, p_max+0.2*dp_fraction)
    p_l_df = numpy.linspace(p_min-0.1*dp_fraction, rp.state_l.p-1e-3*dp_fraction) # want this to go to CJ rather than pmin
    v_l_df = numpy.zeros_like(p_l_df)
    p_l_dt = numpy.linspace(rp.state_l.p+1e-3*dp_fraction, p_max+0.2*dp_fraction) # want this to go from CJ rather than pmax
    v_l_dt = numpy.zeros_like(p_l_dt)
    p_r_1 = numpy.linspace(p_min-0.1*dp_fraction, rp.state_r.p-1e-3*dp_fraction)
    p_r_2 = numpy.linspace(rp.state_r.p+1e-3*dp_fraction, p_max+0.2*dp_fraction)
    p_r_df = numpy.linspace(p_min-0.1*dp_fraction, rp.state_r.p-1e-3*dp_fraction) # want this to go to CJ rather than pmin
    v_r_df = numpy.zeros_like(p_r_df)
    p_r_dt = numpy.linspace(rp.state_r.p+1e-3*dp_fraction, p_max+0.2*dp_fraction) # want this to go from CJ rather than pmax
    v_r_dt = numpy.zeros_like(p_r_dt)

    # rarefaction curves
    v_l_1 = rarefaction(p_l_1, rp.state_l, -1)
    v_r_1 = rarefaction(p_r_1, rp.state_r, 1)

    # shock curves
    v_l_2 = shock(p_l_2, rp.state_l, -1)
    v_r_2 = shock(p_r_2, rp.state_r, 1)

    plot_inert = numpy.ones(4, dtype=numpy.bool)
    # check to make sure there is actually a wave - turn off plotting if not
    if len(rp.waves[0].wave_sections) == 1 and rp.waves[0].wave_sections[0].trivial:
        plot_inert[:2] = False
    if len(rp.waves[-1].wave_sections) == 1 and rp.waves[-1].wave_sections[0].trivial:
        plot_inert[-2:] = False

    # deflagration & detonation curves
    plot_burning = numpy.zeros(4, dtype=numpy.bool)
    if len(rp.waves[0].wave_sections) == 2 or rp.waves[0].wave_sections[0].type == 'Deflagration' or \
        rp.waves[0].wave_sections[0].type == 'Detonation':
        plot_burning[0] = True
        plot_burning[2] = True
        for s in rp.waves[0].wave_sections:
            if s.type == 'Deflagration':
                CJ_v, CJ_p, CJ_rho, CJ_eps = s.q_end.v, s.q_end.p, s.q_end.rho, s.q_end.eps
                p_max=max([p_max, CJ_p])
                CJ_q = s.q_end
                ax.plot(CJ_v, CJ_p, 'ko', label=r"$CJDF$")
                trans_offset = mtrans.offset_copy(ax.transData, fig=fig, x=-15, y=-6, units='dots')
                ax.text(CJ_v, CJ_p, r"$CJ$", transform=trans_offset, horizontalalignment='center',
                verticalalignment='bottom')
                p_l_df = numpy.linspace(CJ_p+1e-3*dp_fraction, rp.state_l.p-1e-3*dp_fraction)
                p_l_dt = numpy.linspace(rp.state_l.p+1e-3*dp_fraction, p_max+0.2*dp_fraction)
                plot_inert[1] = False # get rid of left shock, as this is now a detonation
                # cut off rarefaction at CJ point
                p_l_1 = numpy.linspace(p_min-0.1*dp_fraction, CJ_p-1e-3*dp_fraction)
                v_l_1 = rarefaction(p_l_1, CJ_q, -1)
                v_l_df = deflagration(p_l_df, CJ_q, -1)

            elif s.type == 'Detonation':
                CJ_v, CJ_p, CJ_rho, CJ_eps = s.q_end.v, s.q_end.p, s.q_end.rho, s.q_end.eps
                p_max=max([p_max, CJ_p])
                CJ_q = s.q_end
                ax.plot(CJ_v, CJ_p, 'ko', label=r"$CJDT$")
                trans_offset = mtrans.offset_copy(ax.transData, fig=fig, x=-15, y=-6, units='dots')
                ax.text(CJ_v, CJ_p, r"$CJ$", transform=trans_offset, horizontalalignment='center',
                verticalalignment='bottom')
                p_l_dt = numpy.linspace(rp.state_l.p+1e-3*dp_fraction, CJ_p-1e-3*dp_fraction)
                p_l_df = numpy.linspace(p_min-0.1*dp_fraction, rp.state_l.p-1e-3*dp_fraction)
                plot_inert[0] = False # get rid of left rarefaction - it's now a deflagration
                # cut off shock at CJ point
                p_l_2 = numpy.linspace(CJ_p+1e-3*dp_fraction, p_max+0.2*dp_fraction)
                # redo shock
                v_l_2 = shock(p_l_2, CJ_q, -1)
                v_l_df = deflagration(p_l_df, rp.state_l, -1)

            if s.type == 'Detonation' or s.type == 'Deflagration':
                v_l_dt = detonation(p_l_dt, rp.state_l, -1)

    # right wave
    if len(rp.waves[-1].wave_sections) == 2 or rp.waves[-1].wave_sections[0].type == 'Deflagration' or\
        rp.waves[-1].wave_sections[0].type == 'Detonation':
        plot_burning[1] = True
        plot_burning[3] = True
        for s in rp.waves[-1].wave_sections:

            if s.type == 'Deflagration':
                CJ_v, CJ_p, CJ_rho, CJ_eps = s.q_end.v, s.q_end.p, s.q_end.rho, s.q_end.eps
                CJ_q = s.q_end
                p_max=max([p_max, CJ_p])
                ax.plot(CJ_v, CJ_p, 'ko', label=r"$CJDF$")
                trans_offset = mtrans.offset_copy(ax.transData, fig=fig, x=-15, y=-6, units='dots')
                ax.text(CJ_v, CJ_p, r"$CJ$", transform=trans_offset, horizontalalignment='center',
                verticalalignment='bottom')
                p_r_df = numpy.linspace(CJ_p+1e-3*dp_fraction, rp.state_r.p-1e-3*dp_fraction)
                p_r_dt = numpy.linspace(rp.state_r.p+1e-3*dp_fraction, p_max+0.2*dp_fraction)
                plot_inert[3] = False # get rid of right shock - now a deflagration
                # cut off rarefaction at CJ point
                p_r_1 = numpy.linspace(p_min-0.1*dp_fraction, CJ_p-1e-3*dp_fraction)
                v_r_1 = rarefaction(p_r_1, CJ_q, 1)
                v_r_df = deflagration(p_r_df, CJ_q, 1)

            elif s.type == 'Detonation':
                CJ_v, CJ_p, CJ_rho, CJ_eps = s.q_end.v, s.q_end.p, s.q_end.rho, s.q_end.eps
                CJ_q = s.q_end
                p_max=max([p_max, CJ_p])
                ax.plot(CJ_v, CJ_p, 'ko', label=r"$CJDT$")
                trans_offset = mtrans.offset_copy(ax.transData, fig=fig, x=-15, y=-6, units='dots')
                ax.text(CJ_v, CJ_p, r"$CJ$", transform=trans_offset, horizontalalignment='center',
                verticalalignment='bottom')
                p_r_dt = numpy.linspace(rp.state_r.p+1e-3*dp_fraction, CJ_p-1e-3*dp_fraction)
                p_r_df = numpy.linspace(p_min-0.1*dp_fraction, rp.state_r.p-1e-3*dp_fraction)
                plot_inert[2] = False # get rid of right rarefaction - it's now a deflagration
                # cut off shock at CJ point
                p_r_2 = numpy.linspace(CJ_p+1e-3*dp_fraction, p_max+0.2*dp_fraction)
                v_r_2 = shock(p_r_2, CJ_q, 1)
                v_r_df = rarefaction(p_r_df, rp.state_r, 1)

            if s.type == 'Detonation' or s.type == 'Deflagration':
                v_r_dt = detonation(p_r_dt, rp.state_r, 1)

    # put this here as going to change it for 3 wave case
    ax.set_ylim(0, p_max+0.2*dp_fraction)
    #plt.xlim(v_min-0.5*dv, v_max+0.5*dv)

    # now for weird 3 wave case
    if len(rp.waves[0].wave_sections) == 3:
        #ax.set_ylim(0, p_max+6*dp_fraction)

        # ignition points
        i_p, i_v = rp.waves[0].wave_sections[1].q_start.p, rp.waves[0].wave_sections[1].q_start.v
        ax.plot(i_v, i_p, 'ko')
        trans_offset = mtrans.offset_copy(ax.transData, fig=fig, x=15, y=-6, units='dots')
        ax.text(i_v, i_p, r"$i_S$", transform=trans_offset, horizontalalignment='center',
             verticalalignment='bottom')
        j2, rho, eps, diffp = wave.mass_flux_squared(
            rp.waves[0].wave_sections[1].q_start,
            rp.waves[0].wave_sections[1].q_end.p,
            rp.waves[0].wave_sections[1].q_start.eos)
        v1, v2 = find_pre_ignition(
            rp.waves[0].wave_sections[1].wavespeed,
            j2, rho, -1)
        # choose the one that is closest to other ignition velocity
        if (i_v - v1)**2 < (i_v - v2):
            i_v2 = v1
        else:
            i_v2 = v2

        for s in rp.waves[0].wave_sections:
            # CJ point
            CJ_v, CJ_p, CJ_rho, CJ_eps = s.q_end.v, s.q_end.p, s.q_end.rho, s.q_end.eps
            CJ_q = s.q_end

            p_max=max([p_max, CJ_p, i_p])
            ax.set_ylim(0, p_max+0.2*dp_fraction)

            if s.type == 'Deflagration':
                plot_burning[0] = True
                #CJ_v, CJ_p, CJ_rho, CJ_eps = s.q_end.v, s.q_end.p, s.q_end.rho, s.q_end.eps
                ax.plot(CJ_v, CJ_p, 'ko', label=r"$CJDF$")
                trans_offset = mtrans.offset_copy(ax.transData, fig=fig, x=-15, y=-6, units='dots')
                ax.text(CJ_v, CJ_p, r"$CJ$", transform=trans_offset, horizontalalignment='center',
             verticalalignment='bottom')
                # cut off rarefaction and shock at CJ point
                p_l_1 = numpy.linspace(p_min-0.1*dp_fraction, CJ_p-1e-6*dp_fraction)
                v_l_1 = rarefaction(p_l_1, CJ_q, -1)
                p_l_2 = numpy.linspace(rp.state_l.p+1e-6*dp_fraction, i_p)
                v_l_2 = shock(p_l_2, rp.state_l, -1)
                p_l_df = numpy.linspace(CJ_p+1e-6*dp_fraction, p_max+0.1*dp_fraction)
                v_l_df = rarefaction(p_l_df, CJ_q, -1)

            elif s.type == 'Detonation':
                plot_burning[2] = True
                #CJ_v, CJ_p, CJ_rho, CJ_eps = s.q_end.v, s.q_end.p, s.q_end.rho, s.q_end.eps
                ax.plot(CJ_v, CJ_p, 'ko', label=r"$CJDT$")
                trans_offset = mtrans.offset_copy(ax.transData, fig=fig, x=-15, y=-6, units='dots')
                ax.text(CJ_v, CJ_p, r"$CJ$", transform=trans_offset, horizontalalignment='center',
             verticalalignment='bottom')
                p_l_dt = numpy.linspace(rp.state_l.p+1e-3*dp_fraction, CJ_p+1e-6*dp_fraction)
                v_l_dt = detonation(p_l_dt, rp.state_l, -1)
                # cut off rarefaction and shock at CJ point
                p_l_1 = numpy.linspace(CJ_p+1e-6*dp_fraction, p_max+0.2*dp_fraction)
                v_l_1 = rarefaction(p_l_1, CJ_q, -1)
                p_l_2 = numpy.linspace(p_min-0.1*dp_fraction, CJ_p-1e-6*dp_fraction)
                v_l_2 = shock(p_l_2, CJ_q, -1)

            if s.type == 'Deflagration' or s.type == 'Detonation':
                distance = (i_v2 - v_l_df)**2
                # get approximate ignition point
                i_p2 = p_l_df[numpy.argmin(distance)]
                i_v2 = v_l_df[numpy.argmin(distance)]
                ax.plot(i_v2, i_p2, 'ko')
                trans_offset = mtrans.offset_copy(ax.transData, fig=fig, x=-15, y=-6, units='dots')
                ax.text(i_v2, i_p2, r"$i$", transform=trans_offset, horizontalalignment='center',
                     verticalalignment='bottom')
                # now cut off deflagration
                p_l_df = p_l_df[:numpy.argmin(distance)+1]
                v_l_df = v_l_df[:numpy.argmin(distance)+1]
                ax.plot([i_v, i_v2], [i_p, i_p2], '--k')

                p_max=max([p_max, i_p2])
                ax.set_ylim(0, p_max+0.2*dp_fraction)

    # right wave
    if len(rp.waves[-1].wave_sections) == 3:
         # ignition points
        i_p, i_v = rp.waves[-1].wave_sections[0].q_end.p, rp.waves[-1].wave_sections[0].q_end.v
        ax.plot(i_v, i_p, 'ko')
        trans_offset = mtrans.offset_copy(ax.transData, fig=fig, x=15, y=-6, units='dots')
        ax.text(i_v, i_p, r"$i$", transform=trans_offset, horizontalalignment='center',
             verticalalignment='bottom')
        j2, rho, eps, diffp = wave.mass_flux_squared(
            rp.waves[-1].wave_sections[1].q_start,
            rp.waves[-1].wave_sections[1].q_end.p,
            rp.waves[-1].wave_sections[1].q_start.eos)
        v1, v2 = find_pre_ignition(
            rp.waves[-1].wave_sections[1].wavespeed,
            j2, rho, 1)
        # choose the one that is closest to other ignition velocity
        if (i_v - v1)**2 < (i_v - v2):
            i_v2 = v1
        else:
            i_v2 = v2

        for s in rp.waves[-1].wave_sections:
            CJ_v, CJ_p, CJ_rho, CJ_eps = s.q_end.v, s.q_end.p, s.q_end.rho, s.q_end.eps
            CJ_q = s.q_end
            p_max=max([p_max, CJ_p, i_p])
            ax.set_ylim(0, p_max+0.2*dp_fraction)

            if s.type == 'Deflagration':
                plot_burning[1] = True
                #CJ_v, CJ_p, CJ_rho, CJ_eps = s.q_end.v, s.q_end.p, s.q_end.rho, s.q_end.eps
                ax.plot(CJ_v, CJ_p, 'ko', label=r"$CJDF$")
                trans_offset = mtrans.offset_copy(ax.transData, fig=fig, x=-15, y=-6, units='dots')
                ax.text(CJ_v, CJ_p, r"$CJ$", transform=trans_offset, horizontalalignment='center',
                verticalalignment='bottom')
                # cut off rarefaction at CJ point
                p_r_1 = numpy.linspace(p_min-0.1*dp_fraction, CJ_p-1e-3*dp_fraction)
                v_r_1 = rarefaction(p_r_1, CJ_q, 1)
                p_r_2 = numpy.linspace(rp.state_r.p+1e-3*dp_fraction, i_p)
                v_r_2 = shock(p_r_2, rp.state_r, 1)
                p_r_df = numpy.linspace(CJ_p+1e-6*dp_fraction, p_max+5*dp_fraction)
                v_r_df = deflagration(p_r_df, CJ_q, 1)

            elif s.type == 'Detonation':
                plot_burning[3] = True
                #CJ_v, CJ_p, CJ_rho, CJ_eps = s.q_end.v, s.q_end.p, s.q_end.rho, s.q_end.eps
                ax.plot(CJ_v, CJ_p, 'ko', label=r"$CJDT$")
                trans_offset = mtrans.offset_copy(ax.transData, fig=fig, x=-15, y=-6, units='dots')
                ax.text(CJ_v, CJ_p, r"$CJ$", transform=trans_offset, horizontalalignment='center',
                verticalalignment='bottom')
                # cut off shock at CJ point
                p_r_1 = numpy.linspace(rp.state_r.p+1e-3*dp_fraction, p_max+0.2*dp_fraction)
                v_r_1 = rarefaction(p_r_1, rp.state_r, 1)
                p_r_2 = numpy.linspace(p_min-0.1*dp_fraction, CJ_p-1e-3*dp_fraction)
                v_r_2 = shock(p_r_2, CJ_q, 1)
                p_r_dt = numpy.linspace(CJ_p+1e-6*dp_fraction, p_max+5*dp_fraction)
                v_r_dt = detonation(p_r_dt, CJ_q, 1)

            if s.type == 'Deflagration' or s.type == 'Detonation':
                # find point of intersection of the two curves
                distance = (i_v2 - v_r_df)**2
                # get approximate ignition point
                i_p2 = p_r_df[numpy.argmin(distance)]
                i_v2 = v_r_df[numpy.argmin(distance)]
                ax.plot(i_v2, i_p2, 'ko')
                trans_offset = mtrans.offset_copy(ax.transData, fig=fig, x=-15, y=-6, units='dots')
                ax.text(i_v2, i_p2, r"$i$", transform=trans_offset, horizontalalignment='center',
                verticalalignment='bottom')
                # now cut off deflagration
                p_r_df = p_r_df[:numpy.argmin(distance)+1]
                v_r_df = v_r_df[:numpy.argmin(distance)+1]
                ax.plot([i_v, i_v2], [i_p, i_p2], '--k')

                p_max=max([p_max, i_p2])
                ax.set_ylim(0, p_max+0.2*dp_fraction)


    if plot_inert[0]: ax.plot(v_l_1, p_l_1, '--', label=r"${\cal R}_{\leftarrow}$")
    if plot_inert[1]: ax.plot(v_l_2, p_l_2, '-', label=r"${\cal S}_{\leftarrow}$")
    if plot_inert[2]: ax.plot(v_r_1, p_r_1, '--', label=r"${\cal R}_{\rightarrow}$")
    if plot_inert[3]: ax.plot(v_r_2, p_r_2, '-', label=r"${\cal S}_{\rightarrow}$")
    if plot_burning[0]: ax.plot(v_l_df, p_l_df, '-', label=r"${\cal DF}_{\leftarrow}$")
    if plot_burning[1]: ax.plot(v_r_df, p_r_df, '-', label=r"${\cal DF}_{\rightarrow}$")
    if plot_burning[2]: ax.plot(v_l_dt, p_l_dt, '-', label=r"${\cal DT}_{\leftarrow}$")
    if plot_burning[3]: ax.plot(v_r_dt, p_r_dt, '-', label=r"${\cal DT}_{\rightarrow}$")


    #v_r_1[i] = w_all[-1, var_index]

    #v_r_2[i] = q_end.prim()[var_index]

    if var_invert:
        v_l_1 = 1.0 / v_l_1
        v_r_1 = 1.0 / v_r_1
        v_l_2 = 1.0 / v_l_2
        v_r_2 = 1.0 / v_r_2

    ax.set_xlabel(var_name)
    ax.set_ylabel(r"$p$")
    ax.legend(bbox_to_anchor=(1, 1), loc='upper left')



# -*- coding: utf-8 -*-
"""
Created on Mon Feb 15 11:28:53 2016

@author: ih3
"""

import numpy
from scipy.optimize import brentq
from scipy.integrate import odeint
from copy import deepcopy
from .state import State

def rarefaction_dwdp(w, p, q_known, wavenumber):
    r"""
    There is a tricky point here that needs investigation. If
    the input p is used here, rather than local_state.p, then they
    can diverge (when :math:`v_t` is significant) leading to overflows of g. By
    using local_state we avoid the overflow, but it may mean the final
    state is not very accurate.

    Parameters
    ----------

    w : tuple
        primitive state (rho, v, eps)
    p : scalar
        pressure (required by odeint, but not used: see note above)
    q_known : State
        Known state
    wavenumber : scalar
        Wave number
    """
    lr_sign = wavenumber - 1
    dwdp = numpy.zeros_like(w)
    rho, v, eps = w
    vt = q_known.vt_from_known(rho, v, eps)
    local_state = State(rho, v, vt, eps, q_known.eos)
    cs = local_state.cs
    h = local_state.h
    W_lorentz = local_state.W_lorentz
    xi = local_state.wavespeed(wavenumber)
    # g quantifies the effect of tangential velocities: see the Living Review
    # and original Pons et al paper for details.
    g = vt**2 * (xi**2 - 1.0) / (1.0 - xi * v)**2
    dwdp[0] = 1.0 / (h * cs**2)
    dwdp[1] = lr_sign / (rho * h * W_lorentz**2 * cs) / numpy.sqrt(1.0 + g)
    dwdp[2] = local_state.p / (rho**2 * h * cs**2)
    return dwdp

def mass_flux_squared(q_start, p_end, unknown_eos=None):
    r"""
    Calculates the square of the mass flux through a region, given the state at the start of the region and the pressure at the end.

    Parameters
    ----------

    q_start : State
        State at start of the region
    p_end : scalar
        Pressure at the end of the region
    unknown_eos : dictionary, optional
        Equation of state in the region (provided if different from EoS
        of q_start)
    """

    if unknown_eos is None:
        unknown_eos = q_start.eos

    def shock_root_rho(rho):
        h = unknown_eos['h_from_rho_p'](rho, p_end)
        return (h**2 - q_start.h**2) - \
        (h/rho + q_start.h/q_start.rho) * (p_end - q_start.p)

    if p_end >= q_start.p:
        # Shock
        min_rho = q_start.rho
        shock_root_min = shock_root_rho(min_rho)
        max_rho = numpy.sqrt(p_end/q_start.p) * q_start.rho
        shock_root_max = shock_root_rho(max_rho)
        while(shock_root_min * shock_root_max > 0.0):
            min_rho /= 1.001 # Not sure - could end up with unphysical root?
            max_rho *= 10.0
            shock_root_min = shock_root_rho(min_rho)
            shock_root_max = shock_root_rho(max_rho)
    else:
        # Deflagration
        max_rho = q_start.rho
        shock_root_max = shock_root_rho(max_rho)
        min_rho = numpy.sqrt(p_end/q_start.p) * q_start.rho
        shock_root_min = shock_root_rho(min_rho)
        while(shock_root_min * shock_root_max > 0.0):
            min_rho /= 10.0 # Not sure - could end up with unphysical root?
            max_rho *= 1.001
            shock_root_min = shock_root_rho(min_rho)
            shock_root_max = shock_root_rho(max_rho)
    rho = brentq(shock_root_rho, min_rho, max_rho)
    h = unknown_eos['h_from_rho_p'](rho, p_end)
    eps = h - 1.0 - p_end / rho
    dp = p_end - q_start.p
    dh2 = h**2 - q_start.h**2
    j2 = -dp / (dh2 / dp - 2.0 * q_start.h / q_start.rho)

    return j2, rho, eps, dp

def deflagration_root(p_0_star, q_precursor, unknown_eos, wavenumber, label):
    lr_sign = wavenumber - 1
    j2, rho, eps, dp = mass_flux_squared(q_precursor, p_0_star, unknown_eos)
    if j2 < 0:
        return 10.0 # Unphysical part of Crussard curve, return a random number
    j = numpy.sqrt(j2)
    v_deflagration = (q_precursor.rho**2 *
        q_precursor.W_lorentz**2 * q_precursor.v + \
        lr_sign * j**2 * \
        numpy.sqrt(1.0 + q_precursor.rho**2 *
        q_precursor.W_lorentz**2 *
        (1.0 - q_precursor.v**2) / j**2)) / \
        (q_precursor.rho**2 * q_precursor.W_lorentz**2 + j**2)
    W_lorentz_deflagration = 1.0 / numpy.sqrt(1.0 - v_deflagration**2)
    v = (q_precursor.h * q_precursor.W_lorentz *
         q_precursor.v + lr_sign * dp *
         W_lorentz_deflagration / j) / \
        (q_precursor.h * q_precursor.W_lorentz + dp * (1.0 /
         q_precursor.rho / q_precursor.W_lorentz + \
         lr_sign * q_precursor.v *
         W_lorentz_deflagration / j))
    vt = q_precursor.vt_from_known(rho, v, eps)
    q_unknown = State(rho, v, vt, eps, unknown_eos, label)

    return q_unknown.wavespeed(wavenumber) - v_deflagration

def precursor_root(p_0_star, q_known, wavenumber):
    shock = Shock(q_known, p_0_star, wavenumber)
    q_precursor = shock.q_end
    t_precursor = q_precursor.eos['t_from_rho_eps'](
                    q_precursor.rho, q_precursor.eps)
    t_i = q_precursor.eos['t_ignition'](q_precursor.rho, q_precursor.eps)
    return t_precursor - t_i

# NOTE: all subclasses begin with initialising type, name, wavenumber etc.
#       Can avoid some repeated code by passing these as arguments to
#       superclass constructer and calling that.

# NOTE: To avoid more repeated code: wave speed calculation appears to
#       consist of one or two main parts - a shock wave bit and a burning
#       wave bit. The code for these two sections is almost identical for
#       all subclasses of WaveSection.
#       Could therefore make define functions calculate_shock_speed and
#       calculate_burning_speed for WaveSection class, which are then
#       called by its subclasses

def post_discontinuity_state(p_star, q_start, lr_sign, label, j2, rho, eps, dp,
                             eos_end = None):
    if eos_end is None:
        eos_end = q_start.eos
    j = numpy.sqrt(j2)
    v_shock = (q_start.rho**2 * q_start.W_lorentz**2 * q_start.v + \
        lr_sign * j**2 * \
        numpy.sqrt(1.0 + q_start.rho**2 * q_start.W_lorentz**2 * (1.0 - q_start.v**2) / j**2)) / \
        (q_start.rho**2 * q_start.W_lorentz**2 + j**2)
    W_lorentz_shock = 1.0 / numpy.sqrt(1.0 - v_shock**2)
    v = (q_start.h * q_start.W_lorentz * q_start.v + lr_sign * dp * W_lorentz_shock / j) / \
        (q_start.h * q_start.W_lorentz + dp * (1.0 / q_start.rho / q_start.W_lorentz + \
        lr_sign * q_start.v * W_lorentz_shock / j))
    vt = q_start.vt_from_known(rho, v, eps)
    q_end = State(rho, v, vt, eps, eos_end, label=label)
    return v_shock, q_end

class WaveSection(object):

    def __init__(self, q_start, p_end, wavenumber):
        """
        A part of a wave. For a single shock or rarefaction, this will be the
        complete wave. For a deflagration or detonation, it may form part of
        the full wave.
        """
        # NOTE: what does self.trivial mean?
        self.trivial = False
        assert(wavenumber in [0, 1, 2]), "wavenumber must be 0, 1, 2"
        self.wavenumber = wavenumber
        self.name = None
        self.q_start = None
        self.q_end = None
        self.wavespeed = []
        self.type = ""

    def latex_string(self):
        if self.trivial:
            return ""
        else:
            s = deepcopy(self.name)
            s += r": \lambda^{{({})}}".format(self.wavenumber)
            if len(self.wavespeed) > 1:
                s += r"\in [{:.4f}, {:.4f}]".format(self.wavespeed[0],
                             self.wavespeed[-1])
            else:
                s += r"= {:.4f}".format(self.wavespeed[0])
            return s

    def _repr_latex_(self):
        s = r"$" + self.latex_string() + r"$"
        return s

    def __repr__(self):
        return self.type

    def plotting_data(self):

        if self.trivial:
            data = numpy.zeros((0,8))
            xi = numpy.zeros((0,))
        else:
            data = numpy.vstack((self.q_start.state(), self.q_end.state()))
            xi = numpy.array([self.wavespeed[0], self.wavespeed[0]])

        return xi, data

# NOTE: this class has a different signature to all other subclasses of
#       WaveSection (q_end rather than p_end). Might be more consistent
#       to use the same signature for all subclasses - all could
#       take argument q_end and access variable q_end.p.
class Contact(WaveSection):

    def __init__(self, q_start, q_end, wavenumber):
        """
        A contact.
        """

        self.trivial = False
        assert(wavenumber in [1]), "wavenumber for a Contact must be 1"
        self.type = "Contact"
        self.wavenumber = wavenumber
        self.q_start = deepcopy(q_start)
        self.q_end = deepcopy(q_end)

        self.name = r"{\cal C}"

        self.wavespeed = [q_start.v]

        if numpy.allclose(q_start.state(), q_end.state()):
            self.trivial = True
            self.name = ""

        assert(numpy.allclose(q_start.v, q_end.v)), "Velocities of states "\
        "must match for a contact"
        assert(numpy.allclose(q_start.p, q_end.p)), "Pressures of states "\
        "must match for a contact"
        assert(numpy.allclose(q_start.wavespeed(wavenumber),
                              q_end.wavespeed(wavenumber))), "Wavespeeds of "\
        "states must match for a contact"

class Rarefaction(WaveSection):

    def __init__(self, q_start, p_end, wavenumber):
        """
        A rarefaction.
        """

        self.trivial = False
        assert(wavenumber in [0, 2]), "wavenumber for a Rarefaction "\
        "must be in 0, 2"
        assert(q_start.p >= p_end), "For a rarefaction, p_start >= p_end"
        self.type = "Rarefaction"
        self.wavenumber = wavenumber
        self.q_start = deepcopy(q_start)

        self.name = r"{\cal R}"
        if self.wavenumber == 0:
            label = r"\star_L"
            self.name += r"_{\leftarrow}"
        else:
            label = r"\star_R"
            self.name += r"_{\rightarrow}"

        v_known = q_start.wavespeed(self.wavenumber)

        self.wavespeed = []

        if numpy.allclose(q_start.p, p_end):
            self.trivial = True
            self.q_end = State(q_start.rho, q_start.v, q_start.vt, q_start.eps,
            q_start.eos, label=label)
            v_unknown = v_known
            self.name = ""
        else:
            w_all = odeint(rarefaction_dwdp,
                           numpy.array([q_start.rho, q_start.v, q_start.eps]),
                           [q_start.p, p_end], rtol = 1e-12, atol = 1e-10,
                           args=((q_start, self.wavenumber)))
            self.q_end = State(w_all[-1, 0], w_all[-1, 1],
                              q_start.vt_from_known(w_all[-1, 0], w_all[-1, 1], w_all[-1, 2]),
                              w_all[-1, 2], q_start.eos, label=label)
            v_unknown = self.q_end.wavespeed(self.wavenumber)
            if self.wavenumber == 0:
                self.wavespeed = numpy.array([v_known, v_unknown])
            else:
                self.wavespeed = numpy.array([v_unknown, v_known])

    def plotting_data(self):
        # TODO: make the number of points in the rarefaction plot a parameter
        if self.trivial:
            xi = numpy.zeros((0,))
            data = numpy.zeros((0,8))
        else:
            p = numpy.linspace(self.q_start.p, self.q_end.p, 500)
            w_all = odeint(rarefaction_dwdp,
                           numpy.array([self.q_start.rho,
                               self.q_start.v, self.q_start.eps]),
                               p, rtol = 1e-12, atol = 1e-10,
                               args=(self.q_start, self.wavenumber))
            data = numpy.zeros((len(p),8))
            xi = numpy.zeros_like(p)
            for i in range(len(p)):
                state = State(w_all[i,0], w_all[i,1],
                              self.q_start.vt_from_known(w_all[i,0], w_all[i,1], w_all[i,2]),
                              w_all[i, 2], self.q_start.eos)
                xi[i] = state.wavespeed(self.wavenumber)
                data[i,:] = state.state()

        return xi, data

class Shock(WaveSection):

    def __init__(self, q_start, p_end, wavenumber):
        """
        A shock.
        """

        self.trivial = False
        assert(wavenumber in [0, 2]), "wavenumber for a Shock "\
        "must be in 0, 2"
        # As we use the Shock code for deflagration checks, we can't apply
        # this check
        #assert(q_start.p <= p_end), "For a shock, p_start <= p_end"
        self.type = "Shock"
        self.wavenumber = wavenumber
        lr_sign = self.wavenumber - 1
        self.q_start = deepcopy(q_start)

        self.name = r"{\cal S}"
        if self.wavenumber == 0:
            label = r"\star_L"
            self.name += r"_{\leftarrow}"
        else:
            label = r"\star_R"
            self.name += r"_{\rightarrow}"

        if numpy.allclose(q_start.p, p_end):
            self.trivial = True
            self.q_end = State(q_start.rho, q_start.v, q_start.vt, q_start.eps,
            q_start.eos, label=label)
            v_shock = q_start.wavespeed(self.wavenumber)
            self.name = ""
        else:
            j2, rho, eps, dp = mass_flux_squared(q_start, p_end,
                                                 q_start.eos)
            v_shock, self.q_end = post_discontinuity_state(p_end,
                q_start, lr_sign, label, j2, rho, eps, dp)

        self.wavespeed = [v_shock]

# TODO: Check that q is correctly initialized across each wave in det, defl.
class Deflagration(WaveSection):

    def __init__(self, q_start, p_end, wavenumber):
        """
        A deflagration.
        """

        eos_end = q_start.eos['eos_inert']
        t_i = q_start.eos['t_ignition'](q_start.rho, q_start.eps)

        self.trivial = False
        assert(wavenumber in [0, 2]), "wavenumber for a Deflagration "\
        "must be in 0, 2"
        assert(q_start.p >= p_end), "For a deflagration, p_start >= p_end"
#        t_start = q_start.eos['t_from_rho_eps'](q_start.rho, q_start.eps)
#        assert(t_start >= t_i), "For a deflagration, temperature of start "\
#        "state must be at least the ignition temperature"
        # TODO The above check should be true, but the root-find sometimes just
        # misses. numpy allclose type check?
        self.type = "Deflagration"
        self.wavenumber = wavenumber
        lr_sign = self.wavenumber - 1
        self.q_start = deepcopy(q_start)

        self.name = r"{\cal WDF}"
        if self.wavenumber == 0:
            label = r"\star_L"
            self.name += r"_{\leftarrow}"
        else:
            label = r"\star_R"
            self.name += r"_{\rightarrow}"

        v_known = q_start.wavespeed(self.wavenumber)

        if numpy.allclose(q_start.p, p_end):
            self.trivial = True
            self.q_end = State(q_start.rho, q_start.v, q_start.vt, q_start.eps,
            eos_end, label=label)
            v_deflagration = v_known
            self.name = ""
        else:
            # This is a single deflagration, so the start state must be at the
            # reaction temperature already.
            j2, rho, eps, dp = mass_flux_squared(q_start, p_end, eos_end)
            v_deflagration, q_unknown = post_discontinuity_state(
                p_end, q_start, lr_sign, label, j2, rho, eps, dp,
                eos_end)

            # If the speed in the unknown state means the characteristics are
            # not going into the deflagration, then this is an unstable strong
            # deflagration
            if (lr_sign*(q_unknown.wavespeed(self.wavenumber) - v_deflagration) < 0):
                p_cjdf = brentq(deflagration_root, (1.0+1e-9)*p_end,
                                (1.0-1e-9)*q_start.p,
                                args=(q_start, eos_end, self.wavenumber, label))
                j2, rho, eps, dp = mass_flux_squared(q_start, p_cjdf, eos_end)
                v_deflagration, q_unknown = post_discontinuity_state(
                    p_cjdf, q_start, lr_sign, label, j2, rho, eps,
                    dp, eos_end)
                self.name = r"{\cal CJDF}"
                if self.wavenumber == 0:
                    label = r"\star_L"
                    self.name += r"_{\leftarrow}"
                else:
                    label = r"\star_R"
                    self.name += r"_{\rightarrow}"

            self.q_end = deepcopy(q_unknown)


        self.wavespeed = [v_deflagration]

class Detonation(WaveSection):

    def __init__(self, q_start, p_end, wavenumber):
        """
        A detonation.
        """

        eos_end = q_start.eos['eos_inert']
        t_i = q_start.eos['t_ignition'](q_start.rho, q_start.eps)

        self.trivial = False
        assert(wavenumber in [0, 2]), "wavenumber for a Detonation "\
        "must be in 0, 2"
        assert(q_start.p <= p_end), "For a detonation, p_start <= p_end"
        #t_start = q_start.eos['t_from_rho_eps'](q_start.rho, q_start.eps)
        #assert(t_start >= t_i), "For a detonation, temperature of start "\
        #"state must be at least the ignition temperature"
        self.type = "Detonation"
        self.wavenumber = wavenumber
        lr_sign = self.wavenumber - 1
        self.q_start = deepcopy(q_start)

        self.name = r"{\cal SDT}"
        if self.wavenumber == 0:
            label = r"\star_L"
            self.name += r"_{\leftarrow}"
        else:
            label = r"\star_R"
            self.name += r"_{\rightarrow}"

        v_known = q_start.wavespeed(self.wavenumber)

        if numpy.allclose(q_start.p, p_end):
            self.trivial = True
            self.q_end = State(q_start.rho, q_start.v, q_start.vt, q_start.eps,
            eos_end, label=label)
            v_detonation = v_known
            self.name = ""
        else:
            # This is a single detonation, so the start state must be at the
            # reaction temperature already.
            j2, rho, eps, dp = mass_flux_squared(q_start, p_end, eos_end)
            if j2 < 0:
                # The single detonation is unphysical - must be unstable weak
                # detonation. So skip the calculation and make sure the CJ
                # calculation runs
#                print("Should be a CJ detonation")
                q_unknown = deepcopy(q_start)
                v_detonation = q_unknown.wavespeed(self.wavenumber) + lr_sign
            else:
                v_detonation, q_unknown = post_discontinuity_state(
                    p_end, q_start, lr_sign, label, j2,
                    rho, eps, dp, eos_end)

            # If the speed in the unknown state means the characteristics are
            # not going into the detonation, then this is an unstable weak
            # detonation
            if (lr_sign*(q_unknown.wavespeed(self.wavenumber) - v_detonation) < 0):
                pmin = (1.0+1e-9)*min(q_start.p, p_end)
                pmax = max(q_start.p, p_end)
                fmin = deflagration_root(pmin, q_start, eos_end, self.wavenumber, label)
                fmax = deflagration_root(pmax, q_start, eos_end, self.wavenumber, label)
                while fmin * fmax > 0:
                    pmax *= 2.0
                    fmax = deflagration_root(pmax, q_start, eos_end, self.wavenumber, label)
                p_cjdt = brentq(deflagration_root, pmin, pmax,
                                args=(q_start, eos_end, self.wavenumber, label))
                j2, rho, eps, dp = mass_flux_squared(q_start, p_cjdt, eos_end)
                v_detonation, q_unknown = post_discontinuity_state(
                    p_cjdt, q_start, lr_sign, label, j2, rho,
                    eps, dp, eos_end)
                self.name = r"{\cal CJDT}"
                if self.wavenumber == 0:
                    label = r"\star_L"
                    self.name += r"_{\leftarrow}"
                else:
                    label = r"\star_R"
                    self.name += r"_{\rightarrow}"

            self.q_end = deepcopy(q_unknown)

        self.wavespeed = numpy.array([v_detonation])

def build_inert_wave_section(q_known, unknown_value, wavenumber):
    """
    Object factory for the WaveSection; non-reactive case
    """

    if wavenumber == 1:
        return [Contact(q_known, unknown_value, wavenumber)]
    elif q_known.p < unknown_value:
        return [Shock(q_known, unknown_value, wavenumber)]
    else:
        return [Rarefaction(q_known, unknown_value, wavenumber)]

def build_reactive_wave_section(q_known, unknown_value, wavenumber):
    """
    Object factory for the WaveSection; reactive case
    """

    t_i = q_known.eos['t_ignition'](q_known.rho, q_known.eps)

    if wavenumber == 1:
        return Contact(q_known, unknown_value, wavenumber)
    else:
        wavesections = []
        if q_known.p < unknown_value:
            # The detonation wave
            detonation = Detonation(q_known, unknown_value, wavenumber)
            wavesections.append(detonation)
            q_next = deepcopy(detonation.q_end)
            # Finally, was it a CJ detonation?
            if q_next.p > unknown_value:
                rarefaction = Rarefaction(q_next, unknown_value, wavenumber)
                wavesections.append(rarefaction)
        else:
            t_known = q_known.eos['t_from_rho_eps'](q_known.rho, q_known.eps)
            t_i = q_known.eos['t_ignition'](q_known.rho, q_known.eps)
            if t_known < t_i: # Need a precursor shock
                p_min = unknown_value
                p_max = q_known.p
                t_min = precursor_root(p_min, q_known, wavenumber)
                t_max = precursor_root(p_max, q_known, wavenumber)
                assert(t_min < 0)

                if t_max <= 0:
                    p_max *= 2
                    t_max = precursor_root(p_max, q_known, wavenumber)

                p_0_star = brentq(precursor_root, p_min, p_max,
                                  args=(q_known, wavenumber))
                precursor_shock = Shock(q_known, p_0_star, wavenumber)
                wavesections.append(precursor_shock)
                q_next = precursor_shock.q_end
                q_next.q = q_known.q # No reaction across inert precursor
                q_next.eos = q_known.eos
            else: # No precursor shock
                q_next = deepcopy(q_known)
            # Next, the deflagration wave
            deflagration = Deflagration(q_next, unknown_value, wavenumber)
            wavesections.append(deflagration)
            q_next = deepcopy(deflagration.q_end)
            # Finally, was it a CJ deflagration?
            if q_next.p > unknown_value:
                rarefaction = Rarefaction(q_next, unknown_value, wavenumber)
                wavesections.append(rarefaction)

        return wavesections


class Wave(object):

    def __init__(self, q_known, unknown_value, wavenumber):
        """
        A wave.

        Parameters
        ----------

        self : Wave
            The wave, which has a known state on one side and an unknown
            state on the other side.
        q_known : State
            The known state on one side of the wave
        unknown_value : scalar
            Pressure in the region of unknown state
        wavenumber : scalar
            characterises direction of travel of wave
        """

        # NOTE: it's not so clear what wavenumber is - change to something like a wavedirection variable which can be left/right/static?
        self.wavenumber = wavenumber
        self.wave_sections = []
        self.wavespeed = []

        if 'q_available' not in q_known.eos:
            waves = build_inert_wave_section(q_known, unknown_value,
                                             wavenumber)
            for sections in waves:
                self.wave_sections.append(sections)
        else:
            waves = build_reactive_wave_section(q_known, unknown_value,
                                                wavenumber)
            for sections in waves:
                self.wave_sections.append(sections)

        self.name = self.wave_sections_latex_string()
        if wavenumber == 0:
            self.q_l = deepcopy(q_known)
            if self.wave_sections:
                self.q_r = self.wave_sections[-1].q_end
            else:
                self.q_r = deepcopy(self.q_l)
        elif wavenumber == 1:
            self.q_l = deepcopy(q_known)
            self.q_r = deepcopy(q_known)
        else:
            self.q_r = deepcopy(q_known)
            if self.wave_sections:
                self.q_l = self.wave_sections[-1].q_end
            else:
                self.q_l = deepcopy(self.q_r)

        minspeed = 10
        maxspeed = -10
        if self.wave_sections:
            for wavesection in self.wave_sections:
                for speed in wavesection.wavespeed:
                    minspeed = min(speed, minspeed)
                    maxspeed = max(speed, maxspeed)
        self.wavespeed.append(minspeed)
        if not numpy.allclose(minspeed, maxspeed):
            self.wavespeed.append(maxspeed)

        self.trivial = True
        if self.wave_sections:
            for wavesection in self.wave_sections:
                if not wavesection.trivial:
                    self.trivial = False
        if self.trivial:
            self.wavespeed = []

    def plotting_data(self):

        xi_wave = numpy.zeros((0,))
        data_wave = numpy.zeros((0,8))
        for wavesection in self.wave_sections:
            xi_section, data_section = wavesection.plotting_data()
            xi_wave = numpy.hstack((xi_wave, xi_section))
            data_wave = numpy.vstack((data_wave, data_section))

        if self.wavenumber == 2:
            xi_wave = xi_wave[-1::-1]
            data_wave = data_wave[-1::-1,:]

        return xi_wave, data_wave

    def wave_sections_latex_string(self):
        names = []
        sections = deepcopy(self.wave_sections)
        if self.wavenumber == 2:
            sections.reverse()
        for sec in sections:
            if not sec.trivial:
                names.append(sec.name)
        s = ""
        if len(names)==1:
            s = names[0]
        elif len(names)>1:
            s = r"\left("
            for n in names:
                s += n
            s += r"\right) "
        return s

    def latex_string(self):
        s = self.wave_sections_latex_string()
        speeds = []
        sections = deepcopy(self.wave_sections)
        if self.wavenumber == 2:
            sections.reverse()
        for sec in sections:
            if not sec.trivial:
                for speed in sec.wavespeed:
                    speeds.append(speed)
        if len(speeds) == 0:
            return ""
        elif len(speeds) == 1:
            s += r": \lambda^{{({})}}".format(self.wavenumber)
            s += r"= {:.4f}".format(speeds[0])
        else:
            s += r": \lambda^{{({})}}".format(self.wavenumber)
            s += r"\in [{:.4f}, {:.4f}]".format(min(speeds), max(speeds))

        return s

    def _repr_latex_(self):
        s = r"$" + self.latex_string() + r"$"
        return s
